#include "pch.h"
#include "Racelogic.SetupApiCpp.h"
#include "unmanagedSampleEventArgs.h"
#include <iostream>
#include <fstream>
#include <regex>
#include <sstream>
#include <string>
#include <chrono>
#include <thread>
#include <stdio.h>
#include <vector>

std::ofstream outputFileStream; //used to write sample data to file

//checks if the input parameter matches 'COM' followed by a number
bool isValidComPortString(std::string input)
{
	std::regex regularExpression("COM[0-9]+");
	return std::regex_match(input, regularExpression);
}

//checks if the input parameter matches 'quit' (case insensitive)
bool isQuitString(std::string input)
{
	std::regex regularExpression("[Qq][Uu][Ii][Tt]");
	return std::regex_match(input, regularExpression);
}

//checks if the value matches 'COM' followed by a number or quit (case insensitive)
bool isPortInputValid(std::string userInput)
{
	return isValidComPortString(userInput) || isQuitString(userInput);
}

std::string askUserForCOMPortName()
{
	std::string line;
	do 
	{
		std::cout << "\nPlease enter a COM port to connect to, such as 'COM1' (this is case sensitive), or type 'quit' to quit: ";
		std::getline(std::cin, line);
	} while (!isPortInputValid(line));
	return line;
}

bool connectToVBOX(RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi, std::string portName)
{
	bool success = false;
	int32_t serialNumber;

	std::cout << "Attempting to connect to " << portName << ": ";
	success = SetupApi->Connect(portName, serialNumber);

	if(success)
	{
		std::cout << "Connected\n\nSerial number: " << serialNumber << "\n";
	}
	else
	{
		SetupApi->Disconnect();
		std::cout << "Failed to connect\n";
	}
	return success;
}

std::string extractPortNameFromString(std::string portNameWithDescription)
{
	std::string com = "COM";
	int position = portNameWithDescription.rfind(com);
	int length = portNameWithDescription.length() - (position + 1);
	std::string extractedPort = portNameWithDescription.substr(position, length);

	return extractedPort;
}

bool chooseDeviceAndConnect(RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi)
{
	bool success = false;
	do
	{
		std::string portName;

		portName = askUserForCOMPortName();

		if(!isValidComPortString(portName))
		{
			break;
		}

		success = connectToVBOX(SetupApi, portName);
	} while (!success);
	return success;
}

void listDevices(std::vector<std::string> portNames)
{
	std::cout << "Detected COM devices:\n";

	for (unsigned int i = 0; i < portNames.size(); i++)
	{
		std::cout << portNames[i] << "\n";
	}
}

//list connected devices then either connect automatically if there is just 1, or request user input 
bool listDevicesAndConnect(RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi)
{
	bool success = false;
	std::vector<std::string> portNames;
	portNames = SetupApi->GetPortNames();

	listDevices(portNames);

	switch (portNames.size())
	{
	case 0:
		success = false;
		std::cout << "No COM devices connected. Please ensure a VBOX is connected to your computer.";
		break;
	case 1:
		success = connectToVBOX(SetupApi, extractPortNameFromString(portNames[0]));
		break;
	default:
		success = chooseDeviceAndConnect(SetupApi);
		break;
	}
	return success;
}

void displayFirmwareInformation(RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi)
{
	std::string output = "";

	std::cout << "\nAttempting to retrieve firmware version\n";

	if (SetupApi->GetFirmwareVersion(output))
	{

		std::cout << "Firmware version: " << output << "\n";
	}
	else
	{
		std::cout << "Failed to get firmware version\n";
	}
}

void displayConnectedmodules(RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi)
{
	std::vector<int> serialNumbers;

	if (SetupApi->GetConnectedModules(serialNumbers))
	{
		std::vector<int>::iterator serialNumbersterator;
		std::cout << "\nModule Serial Numbers\n";
		for (serialNumbersterator = serialNumbers.begin(); serialNumbersterator != serialNumbers.end(); serialNumbersterator++)
		{
			std::cout << *serialNumbersterator << ", ";
		}
	}
}

void checkVoltages(RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi)
{
	bool success = false;
	float voltage;
	std::cout << "\nAttempting to retrieve internal battery voltage: ";
	success = SetupApi->GetInternalBatteryVoltage(voltage);
	if (success) 
	{
		std::cout << voltage << "\n";
	}
	else
	{
		std::cout << "Unable to retrieve internal battery voltage\n";
	}

	std::cout << "\nAttempting to retrieve external battery voltage: ";
	success = SetupApi->GetExternalBatteryVoltage(voltage);
	if (success)
	{
		std::cout << voltage << "\n";
	}
	else
	{
		std::cout << "Unable to retrieve external battery voltage\n";
	}
}

//load an rcf configuration file that is compatible with the VBOX Setup application.
void loadConfigurationFile(RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi, std::string fileName)
{
	std::string error;
	std::cout << "\nAttempting to load config file: " << fileName;

	bool success = SetupApi->LoadConfigurationFile(fileName, error);

	if (success)
	{
		std::cout << "\nLoaded config file: " << fileName;
	}
	else
	{
		std::cout << "\nFailed to load config file: " << fileName;
	}
}


//handle sample data received from the VBOX
void* __stdcall onNewSampleMethod(RacelogicSetupApiCpp::UnmanagedSampleEventArgs args)
{
	std::vector<double> data = args.GetData();
	std::vector<double>::iterator iterator;

	outputFileStream << args.GetSatellites() << ",";
	outputFileStream << args.GetUtcTime() << ",";
	outputFileStream << args.GetDgps() << ",";
	outputFileStream << args.GetDualAntenna() << ",";
	outputFileStream << args.GetLatitude() << ",";
	outputFileStream << args.GetLongitude() << ",";
	outputFileStream << args.GetSpeed() << ",";
	outputFileStream << args.GetHeading() << ",";
	outputFileStream << args.GetHeight() << ",";
	outputFileStream << args.GetVerticalVelocity() << ",";
	outputFileStream << args.GetLongitudinalAcceleration() << ",";
	outputFileStream << args.GetLateralAcceleration() << ",";
	outputFileStream << args.GetGlonassSatellites() << ",";
	outputFileStream << args.GetGpsSatellites() << ",";
	outputFileStream << args.GetSerialNumber() << ",";
	outputFileStream << args.GetKalmanFilterCode() << ",";
	outputFileStream << args.GetSolutionType() << ",";
	outputFileStream << args.GetVelocityQuality() << ",";
	outputFileStream << args.GetBrakeTrigger() << ",";
	outputFileStream << args.GetTriggerEventTime() << ",";


	for (iterator = data.begin(); iterator != data.end(); iterator++)
	{
		//output to the file stream
		outputFileStream << *iterator << ",";
	}
	outputFileStream << "\n";

	return 0;
}

//Instruct the VBOX to start outputting sample data
void handleParsing(RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi)
{
	std::cout << "\n\nPress 'Enter'/'Return' to begin reading VBOX data and begin outputting to 'vboxlog.log' (there may be a delay of a few seconds).";

	std::string line;
	std::getline(std::cin, line);

	std::vector<int> serialNumbers;
	std::vector<std::string>channels;
	serialNumbers.clear();
	channels.clear();

	//output to file
	outputFileStream.open("vboxlog.log");

	//start outputting data by calling the onNewSampleMethod each time a new sample is received
	bool success = SetupApi->BeginParse(onNewSampleMethod, channels, serialNumbers);

	if (success)
	{
		std::cout << "\nLogging started.\n";

		std::vector<std::string>::iterator channelsIterator;
		std::cout << "\nChannels being logged\n";
		for (channelsIterator = channels.begin(); channelsIterator != channels.end(); channelsIterator++)
		{
			std::cout << *channelsIterator << ", ";
		}
		std::vector<int>::iterator serialNumbersterator;
		std::cout << "\n\nChannel Serial Numbers\n";
		for (serialNumbersterator = serialNumbers.begin(); serialNumbersterator != serialNumbers.end(); serialNumbersterator++)
		{
			std::cout << *serialNumbersterator << ", ";
		}

		std::cout << "\n\nPress 'Enter'/'Return' to end parsing.\n";
	}
	else
	{
		std::cout << "Failed to start logging. Press 'Enter'/'Return' to continue.\n";
	}

	std::getline(std::cin, line);

	//stop reading data from the VBOX
	SetupApi->EndParse();
	std::this_thread::sleep_for(std::chrono::milliseconds(2000));

	//close the file stream
	outputFileStream.close();
}

//get user confirmation to a yes/no question
bool userWantsTo(std::string questionText)
{
	std::cout << "\n\nDo you want to "<< questionText << "?\ny for yes, n for no: ";

	std::string line;
	std::getline(std::cin, line);
	std::cout << "\n";

	return (line.compare("y") == 0 || line.compare("Y") == 0);
}

std::string getConfigFileNameFromConsole() 
{
	std::this_thread::sleep_for(std::chrono::milliseconds(1000));
	std::string line;
	std::cout << "\nEnter name of, or path to, configuration file:\n";
	std::getline(std::cin, line);

	return line;
}

int main()
{
	std::string line;
	RacelogicSetupApiCpp::RacelogicSetupApiCpp* SetupApi = new RacelogicSetupApiCpp::RacelogicSetupApiCpp();

	while(listDevicesAndConnect(SetupApi))
	{
		bool success = false;;
		displayFirmwareInformation(SetupApi);

		checkVoltages(SetupApi);

		displayConnectedmodules(SetupApi);

		do 
		{
			if(userWantsTo("load a new configuration file"))
			{
				std::string configFileName = getConfigFileNameFromConsole();
				loadConfigurationFile(SetupApi, configFileName);
			}
			handleParsing(SetupApi);
		} while (userWantsTo("carry out another logging session"));
		
		std::cout << "\n\nParsing ended.\nPress 'Enter'/'Return' to disconnect\n";
		std::getline(std::cin, line);

		success = SetupApi->Disconnect();
		if (success)
		{
			std::cout << "Disconnected\n";
		}
		else
		{
			std::cout << "Failed to disconnect\n";
		}
		std::cout << "\n\n";
	}
}
